/*
 * Created on Jul 19, 2005
 *
 */
package com.sunlabs.archsim.assignJan18;

import com.sunlabs.archsim.base.simulation.Delay;
import com.sunlabs.archsim.fleet.common.Data;
import com.sunlabs.archsim.fleet.common.FleetGlobal;
import com.sunlabs.archsim.fleet.common.Payload;
import com.sunlabs.archsim.gasp.Action;
import com.sunlabs.archsim.gasp.Component;
import com.sunlabs.archsim.gasp.Function;
import com.sunlabs.archsim.gasp.Terminal;
import com.sunlabs.archsim.gasp.terminals.InputTerminal;
import com.sunlabs.archsim.gasp.terminals.OutputTerminal;


/*
 * Copyright  2005 Sun Microsystems Inc. All rights reserved. Use is subject to license terms.
 *  
 */

/**
 * This is a basic builidng block for a Fifo. A string of two FifoElement-s
 * forms a one-place fifo, three fifo elements form a two place fifo, ...
 * 
 * Interface: FleetGlobal.IN, FleetGlobal.ACK_IN on the data input side, 
 * and FleetGlobal.OUT, FleetGlobal.ACK_OUT on the output side.
 * 
 * @author ib27688
 */
public class FifoElement extends Component {

	InputTerminal in;
	OutputTerminal ackIn;
	OutputTerminal out;
	InputTerminal ackOut;
	
	FifoAction action;
	
	public FifoElement(String name) {
		super(name, null);
		buildElement();
	}
	
	public FifoElement() {
		// no need for an variable name here
		super("FifoElement", null);		
		buildElement();
	} // FifoElement
	
	protected void buildElement() {
		action= new FifoAction("FifoAction");
		addMember(action);
		
		in= new InputTerminal(FleetGlobal.IN);
		addMember(in);
		action.attach(in);
		
		ackIn= new OutputTerminal(FleetGlobal.ACK_IN);
		addMember(ackIn);
		action.attach(ackIn);
		
		out= new OutputTerminal(FleetGlobal.OUT);
		addMember(out);
		action.attach(out);
		
		ackOut= new InputTerminal(FleetGlobal.ACK_OUT);
		addMember(ackOut);
		action.attach(ackOut);

	}
	
	public Terminal getTerminal(String name) {
		if (name.equals(FleetGlobal.IN)) {
			return in;
		}
		else if (name.equals(FleetGlobal.ACK_IN)) {
			return ackIn;
		}
		else if (name.equals(FleetGlobal.OUT)) {
			return out;
		}
		else if (name.equals(FleetGlobal.ACK_OUT)) {
			return ackOut;
		}
		else
			return null;
	}

	
	public String stateInfo() {
		
		return "FIFOelement " + getName() + ", id= " + getID() + ",  has passed " 
		+ action.counter
		+ " items.";
	} // stateInfo
	
	private class FifoAction extends Action {
		
		int counter= 0;

		public FifoAction(String name) {
			super(name);
			outputFunctions.put(FleetGlobal.OUT, new PassData());
			outputFunctions.put(FleetGlobal.ACK_IN, new ActionFired());
		} // FifoAction
		
		
		/** Simply pass the data from input terminal 0 */
		private class PassData extends Function {
			
			public PassData() {
				super("PassData");
				Integer fDelay= globals.intValue(Assign1Defaults.FIFO_FWD_DELAY);
				functionDelay= (fDelay != null)
					? new Delay(fDelay.intValue())
					: Assign1Defaults.FIFO_FWD_DELAY_DEF;
			}
			
			public Object value() {
				// increase the count of inputs
				counter++;
				// get data from the input
				Object d=  in.getData();
				// is this a payload object?
				if (d instanceof Payload) {
					// get the data item out
					return d;
				}
				else if (d instanceof Data) {
					return d;
				}
				else {
					// wrong input data type, output an illegal value
					director.error("FifoElement " + getAlias() + ", id= " + getID() 
							+ "received data of unexpected data type: " 
							+ d + " of type " + d.getClass().getName()
							);						
					return new Data(d, false);
				}
					
			} // value
		} // PassData

		/** Vacuous function used to acknowldge that the action has fired */ 
		private class ActionFired extends Function {
			
			public ActionFired() {
				super("Vacous");
				Integer fDelay= globals.intValue(Assign1Defaults.FIFO_ACK_DELAY);
				functionDelay= (fDelay != null)
					? new Delay(fDelay.intValue())
					: Assign1Defaults.FIFO_ACK_DELAY_DEF;
			}
		
			public Data value() {
				return null;
			} // value
		} // AcknowledgeData
		
	} // FifoAction
	
} // FifoElement



