package edu.berkeley.sbp;
import edu.berkeley.sbp.util.*;
import edu.berkeley.sbp.*;
import edu.berkeley.sbp.*;
import java.io.*;
import java.util.*;
import java.lang.reflect.*;
import java.lang.ref.*;

/** an element which can produce one of several alternatives */
public class Union extends Element implements Iterable<Sequence> {

    /**
     *  Since every cycle in a non-degenerate grammar contains at
     *  least one Union, every instance of this class must be able to
     *  display itself in both "long form" (list of the long forms of
     *  its alternatives) and "short form" (some abbreviation).
     *
     *  @param shortForm the "short form" display; usually 
     *  @param synthetic if true, this Union's "long form" is "obvious" and should not be displayed when printing the grammar
     */
    public Union() { this(null, false); }
    public Union(String shortForm) { this(shortForm, false); }
    public Union(String shortForm, boolean synthetic) {
        this.shortForm = shortForm;
        this.synthetic = synthetic;
    }

    final String shortForm;
    final boolean synthetic;
    private final List<Sequence> alternatives = new ArrayList<Sequence>();

    public Iterator<Sequence> iterator() { return alternatives.iterator(); }
    public boolean contains(Sequence s) { return alternatives.contains(s); }

    /** adds an alternative */
    public void add(Sequence s) {
        alternatives.add(s);

        // FIXME: does this make sense?
        for(Sequence n : s.needs) add(n);
        for(Sequence n : s.hates) add(n);
    }

    // Epsilon Form //////////////////////////////////////////////////////////////////////////////

    // FIXME
    public static Union epsilon = new Union("()");
    static { epsilon.add(Sequence.empty); }

    // FIXME
    private Forest.Many epsilonForm = null;
    Forest epsilonForm() {
        if (epsilonForm != null) return epsilonForm;
        epsilonForm = new Forest.Many();
        for(Sequence s : this) {
            // FIXME FIXME FIXME
            if (new Walk.Cache().possiblyEpsilon(s))
                epsilonForm.merge(s.epsilonForm());
        }
        return epsilonForm;
    }

    // Display //////////////////////////////////////////////////////////////////////////////

    public String getName() {
        if (shortForm != null) return shortForm;
        return "(anon_union)";
    }
    public String toString() {
        if (shortForm != null) return shortForm;
        StringBuffer sb = new StringBuffer();
        sb.append("(");
        bodyToString(sb, "", " | ");
        sb.append(")");
        return sb.toString();
    }
    public StringBuffer toString(StringBuffer sb) {
        if (synthetic) return sb;
        boolean first = true;
        String before = StringUtil.pad(15, getName()) + " = ";
        if (alternatives.size()==0) {
            sb.append(before);
        } else {
            bodyToString(sb,
                         before,
                         "\n" + StringUtil.pad(15, "")        + " | ");
            sb.append('\n');
        }
        return sb;
    }
    
    private void bodyToString(StringBuffer sb, String before, String between) {
        boolean first = true;
        for(Sequence s : this) {
            if (s.lame) continue;
            // FIXME: what to do here about printing out negated sequences?
            sb.append(first ? before : between);
            first = false;
            sb.append(s.toString());
        }
    }

}
